/* File eventCaptureIO.c, version 1.1, released  09/03/26 at 09:26:29
   retrieved by SCCS 09/03/26 at 09:26:29

%% Program to write captured events to FITS tables

:: Offline diagnostic

HISTORY:
070905 GIL finish subroutine-ization
070821 GIL transform ipcap.c to a subroutine form
070801 P?B initial version in program form.

DESCRIPTION:
eventCaputureIO() reads the latest event from the IBOB
and returns events if found.

DESCRIPTION:
PickIt clone that uses TCP/IP communication to the
iBOB's lwIP low speed ethernet interface.

CHANGES:
* 26 JULY 2007
	- Set initial version at 1.0
	- Runtime measured at 60ms

TODO:
* Add options similar to pickit's
*/

/* Includes used for some general things */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>  /* Time... is on my side... */
#include <time.h>

/* Includes used for most socket operations */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>

/* Includes used for auxiliary calls: memset, etc. */
#include <string.h>
#include <arpa/inet.h>
#include "STDDEFS.H"
#include "EVENT.H"

/* externals */
void tvToMjdUtc(struct timeval *tv, long * mjd, double * utc);

/* Defines */

#define MAX_BUFF_LEN 100000
#define IBOB_ADDR "169.254.128.1"
#define CNT_OFFSET 15
#define PORTNUM 23

/* internals */
int ibobSock = -1;

char * openIbobSocket( int * outSock, struct sockaddr_in * pSin)
{ int sock = 0;
  struct hostent *host;

 * outSock = 0;

  /* Get a socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  /* Check for errors */
  if (sock == -1) {
    perror("socket");
    exit(1);
  }

  /* Clear the structure so that we don't have garbage around */
  memset((void *)pSin, 0, sizeof( struct sockaddr_in));
  
  /* AF means Address Family */
  pSin->sin_family = AF_INET;
  
  /* Fill in port number in address (careful of byte-ordering) */
  pSin->sin_port = htons(PORTNUM);
  
  /* Parse IP */
  if(inet_aton(IBOB_ADDR, &pSin->sin_addr) == 0) {
    host = gethostbyname(IBOB_ADDR);
    pSin->sin_addr = *((struct in_addr *)host->h_addr);
  }

  printf("Connecting to %s:%d ... ", 
	 inet_ntoa(pSin->sin_addr), ntohs(pSin->sin_port));

  *outSock = sock;
  return (NULL);
} /* end of openIbobSocket() */

char * closeIbobSocket()
{ 
  if (ibobSock != -1) 
    close(ibobSock);  /* Close socket */

  ibobSock = -1;
  return (NULL);
} /* end of closeIbobSocket() */

char * getNextEvent( struct EVENTDATA * pEvent)
{ /* Types */
  char buffer[MAX_BUFF_LEN];
  long i = 0, len = 0, count = 0, n = 0, mjd = 0;
  unsigned long eventClk[2], internalClk[2], gpsClk[2], newEvent = FALSE, 
    num_events = 0, temp = 0;
  float float_val = 0;
  double utc = 0;
  static long printCount = 0, isConnected = FALSE;
  static struct sockaddr_in sin;
  struct timeval tv;

  pEvent -> nData = pEvent -> eventNumber = 0;  /* init no new event flag */

  if (ibobSock < 0) 
    openIbobSocket( &ibobSock, &sin);

  if (ibobSock < 0) {
    if (printCount < 10) {
      fprintf( stderr, "Error opening the ibob socket\n");
      printCount++;
    }
    return( "Error opening socket");
  } /* end if error on socket open */

  if (isConnected == FALSE) {
    /* Connect to server (notice the cast and size argument) */
    if(connect( ibobSock, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
      return("Error connecting to Socket");
    }
    else {
      isConnected = TRUE;
      if (printCount < 10) 
	printf("connected.\n");
    }
  } /* end of not yet connected */

  /* To enable normal echoing and command prompt printing: */
  /* send( ibobSock, 255, 1); */

  /* clear any trash from previous calls */
  len = recv( ibobSock, buffer, 1, 100);

  /* Send datadump command */
  strcpy(buffer, "datadump\n");
  if (send( ibobSock, buffer, strlen(buffer), 0) == -1)
    return( "Error Sending datadump request");

  /* Take out \r character */
  memset(buffer, 0, 100);
  len = recv( ibobSock, buffer, 1, 0);
  if(len == -1) 
    return("Error receiving IBOB response");

  /* Get event occurred flag */
  for(i = 0; i < 4; i++) {
    len = recv( ibobSock, &temp, 1, 0);
    if(len == -1) 
      return( "Error getting Event flag");
    newEvent = (newEvent << 8) | (temp & 0xFF);
  }
    
  if (newEvent) { /* in case of new event get rest of data */
    gettimeofday(&tv, 0);
    tvToMjdUtc( &tv, &mjd, &utc);  /* get computer clock time of event */
    pEvent -> mjd = mjd;
    pEvent -> utc = utc;

    /* Get event count */
    for(i = 0; i < 4; i++) {
      len = recv( ibobSock, &temp, 1, 0);
      if (len == -1) 
	return("Error getting the event Count");
      num_events = (num_events << 8) | (temp & 0xFF);
    }
    if (printCount < 10)
      printf("Event Count: %ld\n", num_events);

    pEvent -> eventNumber = num_events;  /* transfer to output */
    
    /* Get sync clock */
    for(i = 0; i < 8; i++) {
      len = recv( ibobSock, &temp, 1, 0);
      if (len == -1) 
	return("Error getting the Sync Clock");
      gpsClk[i/4] = (gpsClk[i/4] << 8) | (temp & 0xFF);
    }
    
    /* Get event clock */
    for (i = 0; i < 8; i++) {
      len = recv( ibobSock, &temp, 1, 0);
      if (len == -1)
	return("Error getting the event Clock");
      eventClk[i/4] = (eventClk[i/4] << 8) | (temp & 0xFF);
    }
      
    /* Get realtime clock */
    for(i = 0; i < 8; i++) {
      len = recv( ibobSock, &temp, 1, 0);
      if (len == -1) 
	return("Error getting the real-time Clock");
      internalClk[i/4] = (internalClk[i/4] << 8) | (temp & 0xFF);
    }

    for (i = 0; i < 2; i++) {
      pEvent -> internalClk[i] = internalClk[i];
      pEvent -> gpsClk[i]      = gpsClk[i];
      pEvent -> eventClk[i]    = eventClk[i];
    }

    /* CNT_OFFSET is used to prevent potential segfaults
       since the while() condition doesn't bounds check */
    count = CNT_OFFSET;
    do {
      /* Get data block */
      len = recv( ibobSock, &buffer[count], MAX_BUFF_LEN, 0);
      if (len == -1) 
	return("Error getting data");
	
      count += len;
      buffer[count] = EOS;
      
    } while(strncmp(&buffer[count-16], "endtransmission", 15) != 0 &&
	    count < MAX_BUFF_LEN);
      
    /* Get rid of the end tag */
    count = count - 17;
      
    /* Write output file header */ 
    if (printCount < 10) {
      printf("Total bytes: %ld\n", count-CNT_OFFSET);
      fprintf( stderr, "# SYN: %lu %lu RT: %lu %lu EVT: %lu %lu\n", 
	       gpsClk[0], gpsClk[1], internalClk[0], internalClk[1], 
	       eventClk[0], eventClk[1]);
      printCount++;
    }

    /* Write data to output file */
    for(i=CNT_OFFSET; i < count; i++) {
      /* Convert from xint8 (8_7) to float */
      float_val = (buffer[i] & 0x7F) / 128.0 - ((buffer[i] >> 7) & 0x01);
	
      /* Insert newline */
      if((i-CNT_OFFSET) % 2 == 1) {
	pEvent->bs[n] = float_val;
	n++;
      }
      else
	pEvent->as[n] = float_val;
    } /* end for all data from IBob */

    pEvent->nData = n;
  } /* end if a new event found */

  return (NULL);
} /* end of eventCaptureIO() */
