/* File ftData.c, version 1.20, released  04/02/18 at 10:57:45 
   retrieved by SCCS 07/07/18 at 17:58:59     

%% GBT spectrometer frequency difference function

:: GBT spectrometer
 
HISTORY:
  031218 GIL limit powerLevel corrections
  030226 GIL check for errors in FFT data, new scheme for mallocs
  020826 GIL document hanning, fix hamming
  020411 GIL use Fred Swabs 3 level code
  020328 GIL limit zeroing to small changes
  020315 GIL fix hamming
  020128 GIL check ave, rms before zeroing offset
  020125 GIL zero offset by averaging
  011129 GIL re-impliment hamming
  011031 GIL use adjust3Level() to get a linear intensity scale
  011012 GIL check for large lag offsets
  011008 GIL do not remove lag bias
  011002 GIL move weighting to magnitude step; fix hanning/hamming weights
  010927 GIL speed up by performing power level correction with weighting
  010925 GIL make function more general; allocate temp arrays
  010919 GIL make function more general; allocate temp arrays
  010816 GIL fix error weighting
  010704 GIL - initial version based on readHagenFile.c

DESCRIPTION:
ftData() and ftData9() do full calibration and Fourier transforming
of 3 and 9 level correllation data.
The input data are assumed to have correllations with range
-1 to 1.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "math.h"
#include "MATHCNST.H"
#include "STDDEFS.H"
#include "transform.h"         /* define transform type indicies */

/* externals */
extern long vanvleck3lev(double *rho, long npts);
extern long vanvleck9lev(double *rho, long npts);
extern char * vanVleck3Fred( double v1in, double lags[], long n);
extern double pow3lev(double zho);
extern double pow9lev(double zho);
extern void four1( float *data, long n, long isign);
extern void realft(float *data, long n, long isign);
extern char * adjust3Level ( double zeroLag, double * gainFactor);
extern double hamming(long index, long npts);
extern char * calcAveRmsRange( double a[], long aStart, long aEnd, 
			       long bStart, long bEnd, 
			       double * aveOut, double * rmsOut);
extern int finite( double);

/* internals */
static long lastYSize = -1;
static long lastWeight = -1;
#define NCOPY 5
#define SMALLOFFSET 999        /* disable offset zeroing */
#define MAXPOWER     11.       /* limit maximum power correction */

char * ftData( long doVanVleck, long doHanning, double yArray[], long n) 
/*ftData() fourier transforms lags to spectra based on weighting and         */
/* smooting parameters.  Returns NULL on OK, else Error messsage             */
{ long i = 0, j = 0;
  double aweight = 1., bweight = 0.,         /* uniform weight */
    powerLevel = 1., zeroLag = 1., dTheta = 0, theta = 0, 
    gainFactor = 1., cosTheta = 1, aveY = 0, rmsY = 0;
  float  * yFloat  = NULL;

  if (n > MAXDATASIZE)
    return("Array too Large to process");

  /* if a change in array size, allocate arrays and print diagnostics */
  if (n != lastYSize || lastWeight != doHanning) {
    fprintf( stderr, "Transform size= %ld; ", n);
    lastYSize = n;
    fprintf( stderr, "Transform Type: ");
    if (doHanning == UNIFORM_WEIGHT)
      fprintf( stderr, "Uniform; ");
    else if (doHanning == HANNING_WEIGHT)
      fprintf( stderr, "Hanning; ");
    else if (doHanning == HAMMING_WEIGHT)
      fprintf( stderr, "Hamming; ");
    else 
      fprintf( stderr, "UNKNOWN; ");
    fprintf( stderr, "VanVleck Type: ");
    if (doVanVleck == RAWVLECK)
      fprintf( stderr, "  Raw   ");
    else if (doVanVleck == VANVLECK3)
      fprintf( stderr, "3 Level ");
    else if (doVanVleck == VANVLECK9)
      fprintf( stderr, "9 Level ");
    else 
      fprintf( stderr, "UNKNOWN ");
    fprintf( stderr, "\n");
    lastWeight = doHanning;
  }

  calcAveRmsRange( yArray, (18*n)/20, (19*n)/20, 0, 0, &aveY, &rmsY);

  /* check for lag offsets, positive or negative */
  if ( aveY > SMALLOFFSET || -aveY > SMALLOFFSET) {
    /* if a significant offset, but not huge */
    if ( (5.*rmsY < aveY || aveY < -5.*rmsY) && (aveY < .1 && -aveY < .1)) {
      for (i = 0; i < n; i++)                /* remove average offset */
        yArray[i] -= aveY;
      fprintf( stderr, "Zeroing large offset in high lags = %e+/-%e\n", 
	       aveY, rmsY);
    } /* end if a significant offset */
  } /* end if offset large */

  zeroLag = yArray[0];

  /* if the input power level is reasonable */
  if (zeroLag != 0.) {
    if (doVanVleck == VANVLECK3) {
      powerLevel = pow3lev( zeroLag);
      /*      vanvleck3lev( yArray, n); */
      vanVleck3Fred( 0., yArray, n);
      /* adjust3Level ( zeroLag, &gainFactor); */
      if (powerLevel > MAXPOWER || powerLevel < .03)
	fprintf (stderr, "Power 3-Level Correction: %f => %f (%f)\n",
	         zeroLag, powerLevel, gainFactor);
    }
    else if (doVanVleck == VANVLECK9) {
      powerLevel = pow9lev( zeroLag);
      vanvleck9lev( yArray, n);

      if (powerLevel > MAXPOWER || powerLevel < .03)
	fprintf (stderr, "Power 9-Level Correction: %f => %f\n",
		 zeroLag, powerLevel);
    } /* else do 9 level correction */
  }
  else
    fprintf( stderr, "Auto-correlation is zero!\n");

  if (powerLevel > MAXPOWER)
    powerLevel = MAXPOWER;

  switch( doHanning) {
    /* Hanning Def:A = 0.5(1-cos(2 pi t/T)) for t = 0 to T; 0 at t=0  */
    /*                                      has a max at t=T */
    /* Hanning Def:A = 0.5(1 - 2*cos^2(pi t/T) + 1) */
    /*             A = 1 - cos^2(pi t/T)            */
    /*             A = sin^2(pi t/T)                */
    case HANNING_WEIGHT:
      aweight = 0.5;
      bweight = 0.5;
      break;
    /* Hamming Def:A = 0.08 + 0.46(1-cos(2 pi t/T))for t = 0 to T; .08 at T=0*/
    /*                                      has a max (1.0) at t=T */
    /*             A = 0.08 + 0.46(1-2*cos^2(pi t/T) +1) */
    /*             A = 0.08 + 0.46(2-2*cos^2(pi t/T)) */
    /*             A = 0.08 + 0.92(sin^2(pi t/T)) */
    case HAMMING_WEIGHT:
      aweight = 0.92;
      bweight = 0.08;
      break;
    default:          /* UNIFORM_WEIGHT  */
      aweight = 1.0;
      bweight = 0.0;
      break;
  }

  /* prepare for hanning and hamming weight */
  dTheta = PI / (2.*n);
  theta = 0.5 * dTheta;

  if (finite( yArray[0]) == 0) { /* if data not finite, warn ! */
    fprintf( stderr, "ftData: Input data array not finite !!!!!!!!\n");
  }

  yFloat  = (float  *)malloc((4*n*sizeof(float))+2);

  /* if uniform weighting */
  if (aweight == 1.0) {
    for(i=0; i<n; i++) {
      yFloat[j] = yArray[i]; 
      j++;
    }
  }
  else if (aweight == 0.5) {               /* else if hanning */
    /* the hanning weight is maximum at zero delay and min at largest delay */
    /* theta goes from 0 to pi/2 so cos(factor) goes 1 -> 0 */
    for(i=0; i<n; i++) {
      cosTheta = cos( theta);
      yFloat[i] = yArray[i] * (cosTheta * cosTheta); 
      theta += dTheta;
    }
  } /* end if weighting the data */
  else {                                  /* else if hamming */
    /* the hanning weight is maximum at zero delay and min at largest delay */
    /* the weight goes from 0.99 to 0.08 */
    for(i=0; i<n; i++) {
      cosTheta = cos( theta);
      yFloat[i] = yArray[i] * (bweight + (aweight*(cosTheta*cosTheta)));
      theta += dTheta;
    }
  } /* end if weighting the data */

  /* now flip order for FT, excluding the zero lag point  */
  j = 2*n - 1;
  for(i=1; i<n; i++) {
    yFloat[j] = yFloat[i];
    j--;
  }

  yFloat[n] = yFloat[n-2];

  /* transfer data to array that is 1 relative, not 0 */
  realft( yFloat-1, 2*n, 1);

  /* now fine tune gain */
  powerLevel *= gainFactor;

  /* compute amplitude from real and imaginary parts */
  j = 0;
  for(i=0; i<n; i++) {
    yArray[i] = powerLevel * yFloat[j];
    j = j + 2;
  }

  free( yFloat);

  if (finite( yArray[0]) == 0) { /* if data not finite, warn ! */
    fprintf( stderr, "ftData: Output data array not finite !!!!!!!!\n");
  }

  return( NULL);
} /* end of ftData() */








