/* File ipcap.c, version 1.3, released  09/04/17 at 14:24:52
   retrieved by SCCS 09/04/20 at 14:31:01

%% Program to retrieve events from an IBOB via 100 Mbps ethernet.

:: Offline diagnostic

HISTORY:
090402 GIL set variance to get 1 event every 10 seconds
090402 GIL add IP address on the command line
090326 GIL add options for 1, 2 or 4 data streams
070802 P?B add ethernet interface
070726 B?R initial version - Runtime measured at 60ms

DESCRIPTION:
PickIt clone that uses TCP/IP communication to the
iBOB's lwIP low speed ethernet interface.

TODO:
* Add options similar to pickit's
*/

/* Includes used for some general things */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>  /* Time... is on my side... */
#include <time.h>

/* Includes used for most socket operations */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>

/* Includes used for auxiliary calls: memset, etc. */
#include <string.h>
#include <arpa/inet.h>

/* Defines */
#define MAX_BUFF_LEN 100000
#define IBOB_ADDR "169.254.128.5"
#define COL_SIZE 9
#define CNT_OFFSET 15
#define L_COL 5
#define R_COL 12
#define PORTNUM 23

/******************************************/
/* This function generates a standardized */
/* date string for use as a filename	  */
/* <YYYY>-<MM>-<DD>T<HH>:<MM>:<SS>.<msms> */
/* Written by Joe Brandt 		  */
/******************************************/
void generate_FITS_date_time_string(struct timeval *tv, char *buf) {
	struct tm tim;

	gmtime_r(&tv->tv_sec, &tim);

	sprintf(buf, "%04d-%02d-%02dT%02d:%02d:%02d.%02d", tim.tm_year+1900, tim.tm_mon+1, tim.tm_mday, tim.tm_hour, tim.tm_min, tim.tm_sec, (tv->tv_usec)/10000);
}


int main(int argc, char *argv[]) {
	/* Types */
	struct timeval tv;
	struct sockaddr_in sin;
	char buffer[MAX_BUFF_LEN] = "", ibobAddress[100] = IBOB_ADDR;
	char filename[40], date[30];
	int i, sock, len, count, total_recv = 0, readCount = 300,
	  variance = 11*16;
	unsigned int evtclk[2], rtclk[2], synclk[2];
	unsigned int new_event, num_events = 0;
	struct hostent *host;
	FILE *output;

        if (argc > 1) {               /* if IP address provided */
	  strcpy( ibobAddress, argv[1]);
	  fprintf( stderr, "Opening Ibob Telnet address %s\n",
		   ibobAddress);
	}

	/* Get a socket */
	sock = socket(AF_INET, SOCK_STREAM, 0);
	/* Check for errors */
	if (sock == -1) {
		perror("socket");
		exit(1);
	}

	/* Clear the structure so that we don't have garbage around */
	memset((void *)&sin, 0, sizeof(sin));

	/* AF means Address Family */
	sin.sin_family = AF_INET;
	
	/* Fill in port number in address (careful of byte-ordering) */
	sin.sin_port = htons(PORTNUM);

	/* Parse IP */
	if(inet_aton( ibobAddress, &sin.sin_addr) == 0) {
		host = gethostbyname(ibobAddress);
		if (host != 0) 
		  sin.sin_addr = *((struct in_addr *)host->h_addr);
		else {
		  fprintf( stderr, "Could Not open Address %s\n", ibobAddress);
		  exit(-1);
		}
	} /* end of test on parsing ibob address */

	printf("Connecting to %s:%d ... ", inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));

	/* Connect to server (notice the cast and size argument) */
	if(connect(sock, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
		perror("connect");
		exit(1);
	}
	else
		printf("connected.\n");

	/* To enable normal echoing and command prompt printing: */
	/* send(sock, 255, 1); */

	while(1) {

	  if ((readCount < 160) && 
	      (readCount % 2 == 0)) {  /* if just starting, test good sigma */
		/* Send datadump command */
		memset(buffer, 0, MAX_BUFF_LEN);
		sprintf(buffer, "regw threshold/variance_i %d \n", variance);
		if (send(sock, buffer, strlen(buffer), 0) == -1) {
			perror("send");
			exit(1);
		}
		printf(buffer);
		sleep(1);
	  } /* end if just starting */

	  readCount ++;
	    
	  /* Send datadump command */
	  memset(buffer, 0, MAX_BUFF_LEN);
	  strcpy(buffer, "datadump\n");
	  if (send(sock, buffer, strlen(buffer), 0) == -1) {
	    perror("send");
	    exit(1);
	  }

	  /* Take out \r character */
	  memset(buffer, 0, strlen(buffer));
	  len = recv(sock, buffer, 1, 0);
	  if(len == -1) {
	    perror("recv");
	    exit(1);
	  }

	  /* Get event occurred flag */
	  for(i = 0; i < 4; i++) {
	    unsigned int temp;
	    len = recv(sock, &temp, 1, 0);
	    if(len == -1) {
	      perror("recv");
	      exit(1);
	    }
	    
	    new_event = (new_event << 8) | (temp & 0xFF);
	  }
	  //printf("Event Occured: %d\n", new_event);

	  if(new_event) {
	    /* Generate file name */
	    gettimeofday(&tv, 0);
	    generate_FITS_date_time_string(&tv, date);
	    
	    printf("Opening output file ... %s.txt\n", date);
	    sprintf(filename, "data/%s.txt", date);
	    output = fopen(filename, "w");
	    if(output == NULL) {
	      printf("Cannot open output file: %s\n", filename);
	      exit(0);
	    }

	    /* Get event count */
	    for(i = 0; i < 4; i++) {
	      unsigned int temp;
	      len = recv(sock, &temp, 1, 0);
	      if(len == -1) {
		perror("recv");
		exit(1);
	      }
	      
	      num_events = (num_events << 8) | (temp & 0xFF);
	    }
	    printf("Event Count: %d\n", num_events);
	    
	    /* Get sync clock */
	    for(i = 0; i < 8; i++) {
	      unsigned int temp;
	      len = recv(sock, &temp, 1, 0);
	      if(len == -1) {
		perror("recv");
		exit(1);
	      }
	      
	      synclk[i/4] = (synclk[i/4] << 8) | (temp & 0xFF);
	    }

	    //Get event clock
	    for(i = 0; i < 8; i++) {
	      unsigned int temp;
	      len = recv(sock, &temp, 1, 0);
	      if(len == -1) {
		perror("recv");
		exit(1);
	      }
	      
	      evtclk[i/4] = (evtclk[i/4] << 8) | (temp & 0xFF);
	    }

			/* Get realtime clock */
			for(i = 0; i < 8; i++) {
				unsigned int temp;
				len = recv(sock, &temp, 1, 0);
				if(len == -1) {
					perror("recv");
					exit(1);
				}

				rtclk[i/4] = (rtclk[i/4] << 8) | (temp & 0xFF);
			}

			/* CNT_OFFSET is used to prevent potential segfaults
			   since the while() condition doesn't bounds check */
			count = CNT_OFFSET;
			//len = recv(sock, buffer, 1, 0); //?
			do {
				/* Get data block */
				len = recv(sock, &buffer[count], MAX_BUFF_LEN, 0);
				if(len == -1) {
					perror("recv");
					exit(1);
				}

				count += len;
				buffer[count] = '\n';
	
			} while(strncmp(&buffer[count-16], "endtransmission\n", 16) != 0);

			/* Get rid of the end tag */
			count = count - 17;

			/* Write output file header */
			printf("Total bytes: %d\n", count-CNT_OFFSET);
			fprintf(output, "# SYN: %u %u RT: %u %u EVT: %u %u\n# DATE: %s\n", synclk[0], synclk[1], rtclk[0], rtclk[1], evtclk[0], evtclk[1], date);

			/* Write data to output file */
			for(i=CNT_OFFSET; i < count; i++) {
				/* Convert from xint8 (8_7) to float */
				float float_val = (buffer[i] & 0x7F) / 128.0 - ((buffer[i] >> 7) & 0x01);

				/* This time stamp only appears once per row */
				if((i-CNT_OFFSET) % 2 == 0) {
					fprintf(output, "%-*d", L_COL, (i-CNT_OFFSET)/2);
				}
	
				/* Print the polarity value to the output file */
				fprintf(output, "%*.6f", R_COL, float_val);

				/* Insert newline */
				if((i-CNT_OFFSET) % 2 == 1) {
					fprintf(output, "\n");
				}
			}

			/* Close file */
			fclose(output);
		} /* end if a new event */
		else {  /* else on new event */
		  if (readCount < 160) {
		    variance = variance - 1;
		    printf ( "Trying variance = %d (%lf)\n", 
			     variance, (double)variance/16.);
		  } /* end of early in the read process */
		} /* end else no new event */
	} /* end of forever */

	/* Close socket */
	close(sock);

	return 0;
}
